import numpy as np
import numpy.linalg as la
import pyopencl as cl
import pyopencl.array as cl_array




KERNEL = """
__kernel void matvec(
  __global const float *a_g,
  __global const float *x_g,
  __global const float *b_g,
  __global float *y_g,
  unsigned int mat_width)
{
    uint i = get_global_id(0);
    uint mat_height = get_global_size(0);

    float result = b_g[i];
    for (int j = 0; j < mat_width; ++j)
        result += a_g[i + j*mat_height] * x_g[j];

    y_g[i] = result;
}
"""




def main():
    ctx = cl.create_some_context()
    queue = cl.CommandQueue(ctx)

    sample_count = 100
    mat_shape = (1024*1024, 32)

    a_host = np.asarray(
            np.random.randn(*mat_shape),
            dtype=np.float32, order="F")
    a_dev = cl_array.to_device(ctx, queue, a_host)

    b_host = np.random.randn(mat_shape[0]).astype(np.float32)
    b_dev = cl_array.to_device(ctx, queue, b_host)

    mat_vec_prg = cl.Program(ctx, KERNEL).build()
    mat_vec_knl = mat_vec_prg.matvec
    mat_vec_knl.set_scalar_arg_dtypes(
            [None, None, None, None, np.uint32])

    norms = []

    print "enter sample loop"

    from time import time
    start_t = time()

    for sample in xrange(sample_count):
        x_host = np.random.uniform(size=mat_shape[1]).astype(np.float32)
        x_dev = cl_array.to_device(ctx, queue, x_host)

        y_dev = cl_array.empty_like(b_dev)
        mat_vec_knl(queue, (mat_shape[0],), (128,), 
                a_dev.data, x_dev.data, b_dev.data, y_dev.data,
                mat_shape[1])
        y_host = y_dev.get()

        #y2_host = np.dot(a_host, x_host)+b_host
        #print la.norm(y_host-y2_host)/la.norm(y2_host)

        norms.append(la.norm(y_host)**2)

    end_t = time()
    elapsed = end_t - start_t

    from pytools import string_histogram
    print string_histogram(np.array(norms)/mat_shape[0], bin_count=30,
            use_unicode=False)

    print

    print "total: %.2e s" % (elapsed/sample_count)




if __name__ == "__main__":
    main()
